﻿using System;
using System.Linq;
using System.Windows.Forms;
using System.Collections.ObjectModel;
using System.Xml.Linq;
using System.Xml;
using System.Collections;
using System.ComponentModel;
/*                           Decision tree with Iris data set                    */
/*                                                                               */
/*                                www.mosismath.com                              */

namespace WindowsFormsApplication1
{ 
    public partial class Mainwin : Form
    {
        const int MAXRECURSIONS = 11;
        const int USELESSLEAVE = 100;
        int samples = 29;
        int dataLength = 0;
        const int features = 4;
        int minClass = 100000;
        int maxClass = -100000;
        int nbOfClasses;
        TData data = new TData();
        BackgroundWorker calcThread;
        TBranch trunc = new TBranch();


        public Mainwin()
        {
            int i,j;
            DataGridViewCell cell;

            InitializeComponent();
            calcThread = new BackgroundWorker();
            calcThread.WorkerReportsProgress = true;
            calcThread.DoWork += worker_Calc;
            calcThread.RunWorkerCompleted += worker_CalculationCompleted;

            data.readData(Application.StartupPath + "\\iris_data.csv");          
            samples = data.values.Count;
            dataLength = data.dataLength;
            dgDataView.RowCount = samples;
            dgDataView.ColumnCount = features + 2;

            dgDataView.Columns[0].HeaderText = "x1";
            dgDataView.Columns[1].HeaderText = "x2";
            dgDataView.Columns[2].HeaderText = "x3";
            dgDataView.Columns[3].HeaderText = "x4";
            dgDataView.Columns[4].HeaderText = "type";
            dgDataView.Columns[5].HeaderText = "output";

            for (i = 0; i < samples; i++)
            {
                dgDataView.Rows[i].HeaderCell.Value = (i + 1).ToString();
                for (j = 0; j <= features; j++)
                {
                    cell = dgDataView[j, i];
                    cell.Value = Math.Round(data.values.ElementAt(i)[j], 4).ToString();
                }
                if (data.values.ElementAt(i)[features] < minClass)
                    minClass = (int)(data.values.ElementAt(i)[features]);
                if (data.values.ElementAt(i)[features] > maxClass)
                    maxClass = (int)(data.values.ElementAt(i)[features]);
            }
            nbOfClasses = maxClass - minClass + 1;
        }

        private void writeSetup(string FileName, TBranch branch)
        {
            XDocument doc = new XDocument();
            doc.Add(new XElement("Treesetting", null));
            try
            {
                using (XmlWriter writer = XmlWriter.Create(FileName))
                {
                    doc.Root.Add(branch.writeBranch(0));
                    doc.WriteTo(writer);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }
        }

        private void readSetup(string FileName, TBranch branch)
        {
            try
            {
                XElement doc = XElement.Load(FileName);
                IEnumerable chields = doc.Elements("Branch");

                foreach (XElement el in chields)
                {
                    branch.loadBranch(el);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }
        }

        private void CalcGini(int iClass, int iAttribute, TBranch branch, Collection<double[]> values)
        {
            bool foundClass = false;
            bool othersEmpty = true;
            double classMin =  1e50;
            double classMax = -1e50;
            double gini;
            int[] iOthersCount = new int[nbOfClasses - 1];
            int[] iOthersClass = new int[nbOfClasses - 1];
            int iActIndex = 0;
            int iClassCount = 0;
            int i, j;
            for (i = 0; i < iOthersClass.Length; i++)
                iOthersClass[i] = -1;          
            for (i = 0; i < values.Count; i++)
            {
                if (values.ElementAt(i)[dataLength - 1] == iClass)
                {
                    foundClass = true;
                    if (values.ElementAt(i)[iAttribute] < classMin)
                        classMin = values.ElementAt(i)[iAttribute];
                    if (values.ElementAt(i)[iAttribute] > classMax)
                        classMax = values.ElementAt(i)[iAttribute];
                }
            }
           
            if (foundClass)
            {
                for (i = 0; i < values.Count; i++)
                {
                    if (values.ElementAt(i)[dataLength - 1] != iClass)
                    {
                        if ((values.ElementAt(i)[iAttribute] > classMax) || (values.ElementAt(i)[iAttribute] < classMin))
                        {
                            foundClass = false;
                            for (j = 0; j < iOthersClass.Length; j++)
                            {
                                if (iOthersClass[j] == (int)values.ElementAt(i)[dataLength - 1])
                                    foundClass = true;
                            }
                            if(!foundClass)
                            {
                                iOthersClass[iActIndex] = (int)values.ElementAt(i)[dataLength - 1];
                                iActIndex++;
                            }
                            for (j = 0; j < iOthersClass.Length; j++)
                            {
                                if (iOthersClass[j] == (int)values.ElementAt(i)[dataLength - 1])
                                    iOthersCount[j]++;
                            }
                        }
                    }
                    else
                        iClassCount++;
                }
                for (j = 0; j < iOthersClass.Length; j++)
                {
                    if (iOthersCount[j] != 0)
                        othersEmpty = false;
                }
                if ((!othersEmpty) || (iClassCount == values.Count))
                {
                    gini = 1.0 - (iClassCount / values.Count) * (iClassCount / values.Count);
                    for (i = 0; i < iOthersClass.Length; i++)
                        gini = gini - (iOthersCount[i] / values.Count) * (iOthersCount[i] / values.Count);
                    branch.purity = gini;
                    branch.min = classMin;
                    branch.max = classMax;
                    branch.iclass = iClass;
                    branch.iProperty = iAttribute;
                }
                else
                    branch.purity = USELESSLEAVE;
            }
            else
                branch.purity = USELESSLEAVE;
        }


        private void BuildBranch(ref TBranch actBranch, Collection<double[]> values)
        {
            int iActClass, iActProperty, j;
         
            for (iActClass = minClass; iActClass <= maxClass; iActClass++)
            {
                for (iActProperty = 0; iActProperty < features; iActProperty++)
                {
                    TBranch branch = new TBranch();
                    CalcGini(iActClass, iActProperty, branch, values);
                    if (actBranch.purity > branch.purity)
                    {
                        actBranch = branch;
                        //actBranch.localValues = values; Just for debugging
                        if (actBranch.purity > 0) // if gini == o there is only one class remaining
                        {
                            Collection<double[]> nextValuesL = new Collection<double[]>();
                            Collection<double[]> nextValuesR = new Collection<double[]>();
                            for (j = 0; j < values.Count; j++)
                            {
                                if ((values.ElementAt(j)[actBranch.iProperty] <= actBranch.max) && (values.ElementAt(j)[actBranch.iProperty] >= actBranch.min))
                                    nextValuesL.Add(values.ElementAt(j));
                                else
                                    nextValuesR.Add(values.ElementAt(j));
                            }
                            if (nextValuesL.Count > 0)
                            {
                                actBranch.nextBranch[0] = new TBranch();
                                BuildBranch(ref actBranch.nextBranch[0], nextValuesL);
                            }
                            if (nextValuesR.Count > 0)
                            {
                                actBranch.nextBranch[1] = new TBranch();
                                BuildBranch(ref actBranch.nextBranch[1], nextValuesR);
                            }
                        }
                    }
                }
            }
        }


        private void worker_Calc(object sender, DoWorkEventArgs e)
        {
            BuildBranch(ref trunc, data.values);
            writeSetup(Application.StartupPath + "\\setup.xml", trunc);
        }

        void worker_CalculationCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            int i, j, flower;
            DataGridViewCell cell;
            for (i = 0; i < data.values.Count; i++)
            {
                flower = trunc.CheckBranch(data.values.ElementAt(i));
                cell = dgDataView[5, i];
                cell.Value = flower.ToString();
            }
            writeSetup(Application.StartupPath + "\\setup.xml", trunc);
            bt_Calc.Enabled = true;
            dgDataView.Enabled = true;
        }

        private void bt_Calc_Click(object sender, EventArgs e)
        {
            bt_Calc.Enabled = false;
            dgDataView.Enabled = false;
            calcThread.RunWorkerAsync();
        }
    }
}
